package flm.b4a.googleplay;

import java.util.ArrayList;
import java.util.List;

import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.Events;
import anywheresoftware.b4a.BA.ShortName;

import com.google.android.gms.common.api.PendingResult;
import com.google.android.gms.common.api.ResultCallback;
import com.google.android.gms.common.api.Status;
import com.google.android.gms.nearby.Nearby;
import com.google.android.gms.nearby.connection.AppIdentifier;
import com.google.android.gms.nearby.connection.AppMetadata;
import com.google.android.gms.nearby.connection.Connections;
import com.google.android.gms.nearby.connection.Connections.ConnectionRequestListener;
import com.google.android.gms.nearby.connection.Connections.EndpointDiscoveryListener;
import com.google.android.gms.nearby.connection.Connections.MessageListener;
import com.google.android.gms.nearby.connection.Connections.StartAdvertisingResult;
import com.google.android.gms.nearby.connection.ConnectionsStatusCodes;

@ShortName("GPlayNearbyConnections")
@Events(values = {
		"onStartAdvertising(ErrorCode As Int, LocalEndpointName As String)",
		"onStartDiscovery(ErrorCode As Int)",
		"onEndpointFound(EndpointId As String, DeviceId As String, AppId As String, EndpointName As String)",
		"onEndpointLost(EndpointId As String)",
		"onSendConnectionRequest(ErrorCode As Int)",
		"onConnectionResponse(ErrorCode As Int, EndpointId As String, Message() As Byte)",
		"onConnectionRequest(EndpointId As String, DeviceId As String, EndpointName As String, Message() As Byte)",
		"onAcceptConnectionRequest(ErrorCode As Int)",
		"onRejectConnectionRequest(ErrorCode As Int)",
		"onMessageReceived(EndpointId As String, Message() As Byte, IsReliable As Boolean)",
		"onDisconnected(EndpointId As String)"
})
public class NearbyConnections implements ConnectionRequestListener, EndpointDiscoveryListener, MessageListener
{
	public final static int STATUS_ALREADY_ADVERTISING = ConnectionsStatusCodes.STATUS_ALREADY_ADVERTISING; 
	public final static int STATUS_ALREADY_CONNECTED_TO_ENDPOINT = ConnectionsStatusCodes.STATUS_ALREADY_CONNECTED_TO_ENDPOINT;
	public final static int STATUS_ALREADY_DISCOVERING = ConnectionsStatusCodes.STATUS_ALREADY_DISCOVERING;
	public final static int STATUS_CONNECTION_REJECTED = ConnectionsStatusCodes.STATUS_CONNECTION_REJECTED; 
	public final static int STATUS_ERROR = ConnectionsStatusCodes.STATUS_ERROR; 
	public final static int STATUS_NETWORK_NOT_CONNECTED = ConnectionsStatusCodes.STATUS_NETWORK_NOT_CONNECTED; 
	public final static int STATUS_NOT_CONNECTED_TO_ENDPOINT = ConnectionsStatusCodes.STATUS_NOT_CONNECTED_TO_ENDPOINT;
	public final static int STATUS_OK = ConnectionsStatusCodes.STATUS_OK;

	public static final long DURATION_INDEFINITE = Connections.DURATION_INDEFINITE;
	public static final int MAX_RELIABLE_MESSAGE_LEN = Connections.MAX_RELIABLE_MESSAGE_LEN;
	public static final int MAX_UNRELIABLE_MESSAGE_LEN = Connections.MAX_UNRELIABLE_MESSAGE_LEN;

	private BA _ba = null;
	private String _evtStartAdvertising, _evtStartDiscovery;
	private String _evtEndpointFound, _evtEndpointLost;
	private String _evtSendConnectionRequest, _evtConnectionResponse;
	private String _evtConnectionRequest;
	private String _evtAcceptConnectionRequest, _evtRejectConnectionRequest;
	private String _evtMessageReceived;
	private String _evtDisconnected;

	/**
	 * Initializes the GPlayNearbyConnections object.
	 * @param EventPrefix Prefix of events.
	 */
	public void Initialize(BA ba, String EventPrefix)
	{
		// Checks whether the Nearby API is connected
		if (!GooglePlayConnection.getAPIclient().hasConnectedApi(Nearby.CONNECTIONS_API))
			throw new RuntimeException("GPlayNearbyConnections.Initialize: nearby connections are not allowed or the client is not connected yet.");

		// Creates the event strings
		String evtPrefix = EventPrefix.toLowerCase(BA.cul);
		_evtStartAdvertising = evtPrefix + "_onstartadvertising";
		_evtStartDiscovery = evtPrefix + "_onstartdiscovery";
		_evtEndpointFound = evtPrefix + "_onendpointfound";
		_evtEndpointLost = evtPrefix + "_onendpointlost";
		_evtSendConnectionRequest = evtPrefix + "_onsendconnectionrequest";
		_evtConnectionResponse = evtPrefix + "_onconnectionresponse";
		_evtConnectionRequest = evtPrefix + "_onconnectionrequest";
		_evtAcceptConnectionRequest = evtPrefix + "_onacceptconnectionrequest";
		_evtRejectConnectionRequest = evtPrefix + "_onrejectconnectionrequest";
		_evtDisconnected = evtPrefix + "_ondisconnected";
		_evtMessageReceived = evtPrefix + "_onmessagereceived";
		_ba = ba;
	}

	public boolean IsInitialized()
	{
		return (_ba != null);
	}

	/** Returns an untranslated debug (not user-friendly!) string based on the specified status code. */
	public static String GetStatusCodeString(int StatusCode)
	{
		return ConnectionsStatusCodes.getStatusCodeString(StatusCode);
	}

	private static String GetStatusCode(int StatusCode)
	{
		return "StatusCode=" + StatusCode + " (" + GetStatusCodeString(StatusCode) + ")";
	}

	/** 
	 * Starts advertising a local endpoint for this app.
	 * @param DurationMillis The duration of the advertisement in milliseconds, unless StopAdvertising is called first. If DURATION_INDEFINITE is passed in, the advertisement will continue indefinitely until StopAdvertising is called.
	 */
	public void StartAdvertising(long DurationMillis)
	{
		// Checks whether the class is initialized
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.StartAdvertising: Duration=" + DurationMillis);
		if (!IsInitialized())
			throw new RuntimeException("StartAdvertising: GPlayNearbyConnections not initialized");

		// Creates the metadata used to describe this application, which can be used to prompt the user to launch or install the application
		List<AppIdentifier> appIdentifierList = new ArrayList<>();
		appIdentifierList.add(new AppIdentifier(BA.packageName));
		AppMetadata appMetadata = new AppMetadata(appIdentifierList);

		// Starts advertising
		PendingResult<StartAdvertisingResult> PR = Nearby.Connections.startAdvertising(GooglePlayConnection.getAPIclient(), null, appMetadata, DurationMillis, this);
		PR.setResultCallback(new ResultCallback<Connections.StartAdvertisingResult>() {
			@Override
			public void onResult(Connections.StartAdvertisingResult result) {
				int statusCode = result.getStatus().getStatusCode();
				if (result.getStatus().isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("StartAdvertising: OK LocalEndpointName=" + result.getLocalEndpointName());
					_ba.raiseEvent(this, _evtStartAdvertising, new Object[] { statusCode, result.getLocalEndpointName() });
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("StartAdvertising: " + GetStatusCode(statusCode));
					_ba.raiseEvent(this, _evtStartAdvertising, new Object[] { statusCode, null });
				}
			}
		});
	}

	/** Stops advertising a local endpoint. Should be called after calling StartAdvertising as soon as the application no longer needs to advertise itself or goes inactive. Messages can still be sent to remote endpoints after advertising ends. */
	public void StopAdvertising()
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.StopAdvertising");
		Nearby.Connections.stopAdvertising(GooglePlayConnection.getAPIclient());
	}

	/** 
	 * Starts discovery for remote endpoints.
	 * @param DurationMillis The duration of discovery in milliseconds, unless StopDiscovery is called first. If DURATION_INDEFINITE is passed in, discovery will continue indefinitely until StopDiscovery is called.
	 */
	public void StartDiscovery(long DurationMillis)
	{
		// Checks whether the class is initialized
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.StartDiscovery: Duration=" + DurationMillis);
		if (!IsInitialized())
			throw new RuntimeException("StartDiscovery: GPlayNearbyConnections not initialized");

		// Starts discovery
		String ServiceID = BA.packageName;
		PendingResult<Status> PR = Nearby.Connections.startDiscovery(GooglePlayConnection.getAPIclient(), ServiceID, DurationMillis, this);
		PR.setResultCallback(new ResultCallback<Status>() {
			@Override
			public void onResult(Status status) {
				int statusCode = status.getStatusCode();
				if (status.isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("StartDiscovery: OK");
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("StartDiscovery: " + GetStatusCode(statusCode));
				}
				_ba.raiseEvent(this, _evtStartDiscovery, new Object[] { statusCode });
			}
		});
	}

	/** Stops discovery for remote endpoints. Should be called after calling StartDiscovery as soon as the client no longer needs to discover endpoints or goes inactive. Messages can still be sent to remote endpoints after discovery ends. */
	public void StopDiscovery()
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.StopDiscovery");
		String ServiceID = BA.packageName;
		Nearby.Connections.stopDiscovery(GooglePlayConnection.getAPIclient(), ServiceID);
	}

	/** Stops advertising and discovery and disconnects from all endpoints. */
	public void StopAllEndPoints()
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.StopAllEndPoints");
		Nearby.Connections.stopAllEndpoints(GooglePlayConnection.getAPIclient());
	}

	/** Returns the ID of the device, used when communicating with other devices. This identifier will be the same for all clients on this device, and will be stable across reboots of the device. */
	public String getLocalDeviceId()
	{
		return Nearby.Connections.getLocalDeviceId(GooglePlayConnection.getAPIclient());
	}

	/** Returns the ID of the local endpoint, used when communicating with other devices. This identifier will be different for each Google API client. */
	public String getLocalEndpointId()
	{
		return Nearby.Connections.getLocalEndpointId(GooglePlayConnection.getAPIclient());
	}

	/**
	 * Sends a request to connect to a remote endpoint.
	 * @param RemoteEndpointId The identifier for the remote endpoint to which a connection request will be sent. Should match the value provided by the onEndpointFound event.
	 * @param Message Bytes of a custom message to send with the connection request. This message must not exceed MAX_RELIABLE_MESSAGE_LEN bytes in length. May be Null.
	 */
	public void SendConnectionRequest(String RemoteEndpointId, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.SendConnectionRequest: " + RemoteEndpointId);
		Connections.ConnectionResponseCallback Response = new Connections.ConnectionResponseCallback() {
			@Override
			public void onConnectionResponse(String endpointId, Status status, byte[] bytes) {
				int statusCode = status.getStatusCode();
				if (status.isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("onConnectionResponse: OK " + endpointId);
					_ba.raiseEvent(this, _evtConnectionResponse, new Object[] { statusCode, endpointId, bytes });
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("onConnectionResponse: " + GetStatusCode(statusCode) + " " + endpointId);
					_ba.raiseEvent(this, _evtConnectionResponse, new Object[] { statusCode, endpointId, null });
				}
			}
		};
		PendingResult<Status> PR = Nearby.Connections.sendConnectionRequest(GooglePlayConnection.getAPIclient(), null, RemoteEndpointId, Message, Response, this);
		PR.setResultCallback(new ResultCallback<Status>() {
			@Override
			public void onResult(Status status) {
				int statusCode = status.getStatusCode();
				if (status.isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("SendConnectionRequest: OK");
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("SendConnectionRequest: " + GetStatusCode(statusCode));
				}
				_ba.raiseEvent(this, _evtSendConnectionRequest, new Object[] { statusCode });
			}
		});
	}

	/**
	 * Accepts a connection request from a remote endpoint. This method must be called before messages can be received from the remote endpoint.
	 * @param RemoteEndpointId The identifier for the remote endpoint that sent the connection request. Should match the value provided by the onConnectionRequest event.
	 * @param Message Bytes of a custom message to send with the connection response. This message must not exceed MAX_RELIABLE_MESSAGE_LEN bytes in length. May be Null.
	 */
	public void AcceptConnectionRequest(String RemoteEndpointId, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.AcceptConnectionRequest: " + RemoteEndpointId);
		PendingResult<Status> PR = Nearby.Connections.acceptConnectionRequest(GooglePlayConnection.getAPIclient(), RemoteEndpointId, Message, this);
		PR.setResultCallback(new ResultCallback<Status>() {
			@Override
			public void onResult(Status status) {
				int statusCode = status.getStatusCode();
				if (status.isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("AcceptConnectionRequest: OK");
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("AcceptConnectionRequest: " + GetStatusCode(statusCode));
				}
				_ba.raiseEvent(this, _evtAcceptConnectionRequest, new Object[] { statusCode });
			}
		});
	}

	/**
	 * Rejects a connection request from a remote endpoint.
	 * @param RemoteEndpointId The identifier for the remote endpoint that sent the connection request. Should match the value provided by the onConnectionRequest event.
	 */
	public void RejectConnectionRequest(String RemoteEndpointId)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.RejectConnectionRequest: " + RemoteEndpointId);
		PendingResult<Status> PR = Nearby.Connections.rejectConnectionRequest(GooglePlayConnection.getAPIclient(), RemoteEndpointId);
		PR.setResultCallback(new ResultCallback<Status>() {
			@Override
			public void onResult(Status status) {
				int statusCode = status.getStatusCode();
				if (status.isSuccess()) {
					// Success !
					if (GooglePlayConnection.DEBUG) BA.Log("RejectConnectionRequest: OK");
				} else {
					// Failure
					if (GooglePlayConnection.DEBUG) BA.Log("RejectConnectionRequest: " + GetStatusCode(statusCode));
				}
				_ba.raiseEvent(this, _evtRejectConnectionRequest, new Object[] { statusCode });
			}
		});
	}

	@BA.Hide
	@Override
	public void onConnectionRequest(final String endpointId, String deviceId, String endpointName, byte[] payload)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.onConnectionRequest = " + endpointId + " " + deviceId + " " + endpointName);
		_ba.raiseEvent(this, _evtConnectionRequest, new Object[] { endpointId, deviceId, endpointName, payload });
	}

	@BA.Hide
	@Override
	public void onEndpointFound(final String endpointId, String deviceId, String serviceId, final String endpointName)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.onEndpointFound: " + endpointId + " " + deviceId + " " + serviceId + " " + endpointName);
		_ba.raiseEvent(this, _evtEndpointFound, new Object[] { endpointId, deviceId, serviceId, endpointName });
	}

	@BA.Hide
	@Override
	public void onEndpointLost(String endpointId)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.onEndpointLost: EndpointId=" + endpointId);
		_ba.raiseEvent(this, _evtEndpointLost, new Object[] { endpointId });
	}

	/**
	 * Sends a message to a remote endpoint using a reliable protocol. Reliable messages will be retried until delivered, and are delivered in the order they were sent to a given endpoint. Messages can only be sent to remote endpoints once a connection request was first sent and accepted (in either direction).
	 * @param RemoteEndpointId The identifier for the remote endpoint to which the message should be sent.
	 * @param Message The bytes of the message to send to the remote endpoint. This message must not exceed MAX_RELIABLE_MESSAGE_LEN bytes in length.
	 */
	public void SendReliableMessage(String RemoteEndpointId, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.SendReliableMessage");
		Nearby.Connections.sendReliableMessage(GooglePlayConnection.getAPIclient(), RemoteEndpointId, Message);
	}

	/**
	 * Sends a message to a list of remote endpoints using a reliable protocol. Reliable messages will be retried until delivered, and are delivered in the order they were sent to a given endpoint. Messages can only be sent to remote endpoints once a connection request was first sent and accepted (in either direction).
	 * @param RemoteEndpointIds The list of identifiers for the remote endpoints to which the message should be sent.
	 * @param Message The bytes of the message to send to the remote endpoints. This message must not exceed MAX_RELIABLE_MESSAGE_LEN bytes in length.
	 */
	public void SendReliableMessage2(anywheresoftware.b4a.objects.collections.List RemoteEndpointIds, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.SendReliableMessage2");
		ArrayList<String> EndpointsList = Utils.FromB4AListToArrayList_String(RemoteEndpointIds);
		Nearby.Connections.sendReliableMessage(GooglePlayConnection.getAPIclient(), EndpointsList, Message);
	}

	/**
	 * Sends a message to a remote endpoint using an unreliable protocol. Unreliable messages may be dropped or delivered out of order. Messages can only be sent to remote endpoints once a connection request was first sent and accepted (in either direction).
	 * @param RemoteEndpointId The identifier for the remote endpoint to which the message should be sent.
	 * @param Message The bytes of the message to send to the remote endpoint. This message must not exceed MAX_UNRELIABLE_MESSAGE_LEN bytes in length.
	 */
	public void SendUnreliableMessage(String RemoteEndpointId, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.SendUnreliableMessage");
		Nearby.Connections.sendUnreliableMessage(GooglePlayConnection.getAPIclient(), RemoteEndpointId, Message);
	}

	/**
	 * Sends a message to a list of remote endpoints using an unreliable protocol. Unreliable messages may be dropped or delivered out of order. Messages can only be sent to remote endpoints once a connection request was first sent and accepted (in either direction).
	 * @param RemoteEndpointIds The list of identifiers for the remote endpoints to which the message should be sent.
	 * @param Message The bytes of the message to send to the remote endpoints. This message must not exceed MAX_UNRELIABLE_MESSAGE_LEN bytes in length.
	 */
	public void SendUnreliableMessage2(anywheresoftware.b4a.objects.collections.List RemoteEndpointIds, byte[] Message)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.SendUnreliableMessage2");
		ArrayList<String> EndpointsList = Utils.FromB4AListToArrayList_String(RemoteEndpointIds);
		Nearby.Connections.sendUnreliableMessage(GooglePlayConnection.getAPIclient(), EndpointsList, Message);
	}

	@BA.Hide
	@Override
	public void onMessageReceived(String endpointId, byte[] payload, boolean isReliable)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.onMessageReceived: EndpointId=" + endpointId + " IsReliable=" + isReliable);
		_ba.raiseEvent(this, _evtMessageReceived, new Object[] { endpointId, payload, isReliable });
	}

	/** Disconnects from a remote endpoint. Messages can no longer be sent to or received from the endpoint after this method is called. */
	public void DisconnectFromEndpoint(String RemoteEndpointId)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.DisconnectFromEndpoint: " + RemoteEndpointId);
		Nearby.Connections.disconnectFromEndpoint(GooglePlayConnection.getAPIclient(), RemoteEndpointId);
	}

	@BA.Hide
	@Override
	public void onDisconnected(String endpointId)
	{
		if (GooglePlayConnection.DEBUG) BA.Log("GPlayNearbyConnections.onDisconnected: EndpointId=" + endpointId);
		_ba.raiseEvent(this, _evtDisconnected, new Object[] { endpointId });
	}
}
