package flm.b4a.googleplay;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.Permissions;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.objects.collections.Map;

@ShortName("GPlayNetworkInfo")
@BA.Events(values = {
		"onConnectivityChange(NetworkType As Int, IsConnectedOrConnecting As Boolean)"
})
@Permissions(values = { "android.permission.ACCESS_NETWORK_STATE" })
public class Network_Info
{
	public static final int NETWORK_NONE = -1;
	public static final int NETWORK_MOBILE = ConnectivityManager.TYPE_MOBILE;
	public static final int NETWORK_WIFI = ConnectivityManager.TYPE_WIFI;
	public static final int NETWORK_WIMAX = 6;
	public static final int NETWORK_BLUETOOTH = 7;
	public static final int NETWORK_ETHERNET = 9;
	public static final int NETWORK_VPN = 17;

	private static ConnectivityManager _CM = null;
	private static ConnectivityChangeReceiver _CNR = null;

	public void Initialize()
	{
		_CM = (ConnectivityManager) BA.applicationContext.getSystemService(Context.CONNECTIVITY_SERVICE);
	}

	public boolean IsInitialized()
	{
		return (_CM != null);
	}

	/** Returns a map of all available networks. In the map, key = type of network (one of the NETWORK constants), value = whether it is connected. */
	public Map getAllNetworkInfo()
	{
		NetworkInfo[] Info = _CM.getAllNetworkInfo();
		Map mapNetworks = new Map();
		mapNetworks.Initialize();
		for (NetworkInfo ni : Info) {
			if (ni.isAvailable())
				mapNetworks.Put(ni.getType(), ni.isConnected());  
		}
		return mapNetworks;
	}

	/** Returns the type of the network in the process of establishing a connection (one of the NETWORK constants). */
	public int getConnectingNetwork()
	{
		NetworkInfo[] Info = _CM.getAllNetworkInfo();
		for (NetworkInfo ni : Info) {
			if (ni.isConnectedOrConnecting())
				return ni.getType();  
		}
		return NETWORK_NONE;
	}

	/** Returns the type of the connected network (one of the NETWORK constants). */
	public int getConnectedNetwork()
	{
		NetworkInfo[] Info = _CM.getAllNetworkInfo();
		for (NetworkInfo ni : Info) {
			if (ni.isConnected())
				return ni.getType();  
		}
		return NETWORK_NONE;
	}

	/** Returns the type of the currently active default data network (one of the NETWORK constants). */
	public int getActiveNetwork()
	{
		NetworkInfo ni = _CM.getActiveNetworkInfo();
		if (ni == null)
			return NETWORK_NONE;
		else
			return ni.getType();
	}

	/** Returns whether the currently active default data network is connected. */
	public boolean IsConnected()
	{
		NetworkInfo ni = _CM.getActiveNetworkInfo();
		if (ni == null)
			return false;
		else
			return ni.isConnected();
	}

	/** Indicates whether the current attempt to connect to the network resulted from the connectivity manager trying to fail over to this network following a disconnect from another network. */
	public boolean IsFailover()
	{
		NetworkInfo ni = _CM.getActiveNetworkInfo();
		if (ni == null)
			return false;
		else
			return ni.isFailover();
	}

	/** Indicates whether the device is currently roaming on the given network. */
	public boolean IsRoaming(int NetworkType)
	{
		NetworkInfo ni = _CM.getNetworkInfo(NetworkType);
		if (ni == null)
			return false;
		else
			return ni.isRoaming();
	}

	/**
	 *Registers a listener to be notified of connectivy changes.
	 *The onConnectivityChange events will return the type of network (one of the NETWORK constants) affected by the change and whether this network is connected or not.
	 */
	public static void RegisterConnectivityChangeListener(BA ba)
	{
		_CNR = new ConnectivityChangeReceiver(ba);
		ba.context.registerReceiver(_CNR, new IntentFilter(ConnectivityManager.CONNECTIVITY_ACTION));
	}

	/** Stops the notification of connectivy changes. */
	public static void UnregisterConnectivityChangeListener(BA ba)
	{
		if (_CNR != null) {
			ba.context.unregisterReceiver(_CNR);
			_CNR = null;
		}
	}

	private static class ConnectivityChangeReceiver extends BroadcastReceiver
	{
		private BA _ba; 
		private String _evtName;

		public ConnectivityChangeReceiver(BA ba)
		{
			_ba = ba;
			_evtName = GooglePlayConnection._evtPrefix + "_onconnectivitychange";
		}

		@Override
		public void onReceive(Context context, Intent intent)
		{
			NetworkInfo ni = _CM.getActiveNetworkInfo();
			int Type;
			if (ni == null) {
				ni = intent.getParcelableExtra(ConnectivityManager.EXTRA_NETWORK_INFO);
				Type = NETWORK_NONE;
			}
			else
				Type = ni.getType();
			if (GooglePlayConnection.DEBUG) BA.Log("NetworkInfo.onConnectivityChange: " + ni.toString());
			_ba.raiseEvent2(this, true, _evtName, false,
					new Object[] { Type, ni.isConnectedOrConnecting() });
		}
	}
}
